/* FILE: director.h                 -*-Mode: c++-*-
 *
 * 3D solver director class.
 *
 */

#ifndef _OXS_DIRECTOR
#define _OXS_DIRECTOR

#include <map>
#include <string>
#include <vector>

#include "oc.h"
#include "key.h"
#include "simstate.h"
#include "mesh.h"
#include "meshvalue.h"
#include "threevector.h"
#include "output.h"
#include "energy.h"	// Needed by MSVC++ 5

OC_USE_STD_NAMESPACE;  // Specify std namespace, if supported

/* End includes */

class Oxs_Ext;       // Forward declarations
class Oxs_Energy;
class Oxs_Driver;

// 3D solver director
class Oxs_Director {
private:
  static void ExitProc(ClientData);

  UINT4m problem_count; // Number of problems loaded across history
  /// of this instance.  Initialized to 0, then incremented to 1 when
  /// first problem is loaded, and incremented by 1 more each time
  /// a new problem is loaded.

  UINT4m problem_id; // Current problem count id.  Set to 0
  /// if not problem is zero.  Otherwise should equal problem_count.
  /// NB: ProbRun is disabled if problem_id==0.

  unsigned int random_number_seed;  // Oc_UnifRand() is called with
  /// this value before Oxs_Ext object creation, and again before
  /// Oxs_Ext Init() calls.  If non-zero, then this seed is also
  /// sent to srand in safe_interp.

  UINT4m energy_calc_count; // Number of times GetEnergyDensity
  /// has been called in current problem run.

  Tcl_Interp* const interp;

  // Safe interpreter, which is a slave of interp.  This is
  // created inside ProbInit(), and is held until Release()
  Tcl_Interp* safe_interp;
  string safe_interp_name;

  void Release();

  vector<Oxs_Ext*> ext_obj;  // Ordered list of instantiated Oxs_Ext
  /// objects.  This list "owns" the objects, and is responsible for
  /// their destruction.
  map<string,Oxs_Ext*> ext_map; // Mapping of Ext names to objects
  vector<Oxs_Energy*> energy_obj; // Ordered list of energy terms.  This
  // is a sublist of ext_obj.

  vector<Oxs_Output*> output_obj; // List of all registered output objects
  const char* MakeOutputToken(unsigned long index,unsigned long probid) const;
  Oxs_Output* InterpretOutputToken(const char* token) const;

  Oxs_Driver* driver; // Unique driver object.  This can be changed
  /// to a container if in the future we want to support multiple
  /// drivers in one problem run.

  // Array of held simulation states.  See discussion below concerning
  // ReserveSimulationStateRequest() and GetNewSimulationState().  This
  // array "owns" the simulations states, i.e., is responsible for their
  // destruction.
  UINT4m simulation_state_reserve_count;
  vector<Oxs_SimState*> simulation_state;

  // NB: Because ext_map and energy_obj hold duplicate pointers to
  // objects stored in ext_obj, one has to be careful to release all
  // pointers at the same time.  This is handled inside the Release()
  // function.  We could store Oxs_Keys in ext_map and energy_obj
  // instead of pointers; this would provide notification of hanging
  // pointers, but since Oxs_Director is the owner of these objects, at
  // present I don't think this is necessary. -mjd

  mutable Oxs_MeshValue<REAL8m> temp_energy;
  mutable Oxs_MeshValue<ThreeVector> temp_field;

  // Outputs maintained by the director.  These are conceptually
  // public, but are specified private to force clients to use the
  // output_map interface.
  void UpdateOutputs(const Oxs_SimState&);
  void FillEnergyCalcCountOutput(const Oxs_SimState&);
  Oxs_ScalarOutput<Oxs_Director> total_energy_output;
  Oxs_VectorFieldOutput<Oxs_Director> total_field_output;
  Oxs_ScalarOutput<Oxs_Director> energy_calc_count_output;

public:
  Oxs_Director(Tcl_Interp* interp_);
  ~Oxs_Director();  // Non-virtual.  Oxs_Director is not intended to
  // have children.

  int VerifyInterp(Tcl_Interp* i) { return (i==interp ? 1 : 0); }

  int ProbInit(const char* mif,const char* filename);
  /// Import mif must be the name of an existing MIF object
  /// in the Tcl interpreter 'interp' passed to the constructor
  /// of *this.
  int ProbReset();
  int ProbRelease() { Release(); return TCL_OK; }
  int ProbRun(); // Passes control over to driver
  int ProbRun(Tcl_Interp* interp, int argc, char** argv);

  Oxs_Ext* FindExtObject(const string& id) const;
  /// Returns a pointer to the element of ext_map with instance id
  /// 'id'.  Returns NULL if the requested object is not
  /// found.  For convenience, some partial matches are allowed.  See
  /// the routine implementation for details.
  ///    It is recommended that the client place a dep lock on the
  /// object if it is to be used for an extended period.  All such locks
  /// must be released before the corresponding object in ext_map can be
  /// deleted.  This means that locks held by ext_map objects to other
  /// ext_map objects should be only to objects earlier in the ext_obj
  /// list than themselves.  Perhaps there should be some type of access
  /// control on this function?

  string ListExtObjects() const;  // Mainly for debugging.
  string ListEnergyObjects() const;  // Mainly for debugging.

  // The following output (de)registration routines throw exceptions
  // on failure. NOTE: All objects registering output are expected to
  // automatically deregister as a consequence of the object destructions
  // occuring in Oxs_Director::Release().  See also output_map object
  // above.
  void RegisterOutput(Oxs_Output*);
  void DeregisterOutput(const Oxs_Output*);
  BOOL IsRegisteredOutput(const Oxs_Output*) const;
  void ListOutputObjects(vector<string> &outputs) const;
  int Output(const char* output_token,
	     Tcl_Interp* interp,
	     int argc,char** argv);
  int OutputCacheRequestIncrement(const char* output_token,
				  INT4m incr) const;
  int OutputNames(const char* output_token,
		  string& owner_name,string& output_name,
		  string& output_type,string& output_units) const;

  Oxs_Output* FindOutputObject(const char* regexp) const;
  /// C++ interface for searching the output object list.  This routine
  /// matches the import regexp against the output objects list.  If a
  /// unique match is found, then a pointer to that object is returned.
  /// Otherwise, a NULL pointer is returned.  The match is performed
  /// using the Tcl regexp engine.  Since this routine will typically
  /// check through the entire list of output objects, ideally this
  /// routine is called after problem initialization, and the result is
  /// cached until the problem is released.


  void GetEnergyDensity(const Oxs_SimState& state,
			Oxs_MeshValue<REAL8m>& energy,
			Oxs_MeshValue<ThreeVector>* mxH_req,
			Oxs_MeshValue<ThreeVector>* H_req
			);
  /// Note that mxH is returned, as opposed to MxH.  This
  /// relieves this routine from needing to know what Ms is,
  /// and saves an unneeded multiplication (since the evolver
  /// is just going to divide it back out again to calculate
  /// dm/dt (as opposed again to dM/dt)).  The returned energy
  /// array is average energy density for the corresponding cell
  /// in J/m^3; mxH is in A/m.  Any of mxH or H may be NULL,
  /// which disables assignment for that variable.

  // SIMULATION STATE MANAGEMENT: During construction, each Oxs_Ext
  // object that wants to create and/or hold simulation states should
  // register its intent with the director using
  // ReserveSimulationStateRequest().  The states are actually
  // constructed only as needed, but the count obtained via reserve
  // calls is used as an upper bound to protect against memory leaks
  // swallowing up all of system memory.  Objects obtain keys (handles)
  // to new states with the GetNewSimulationState() call.  The key
  // initially holds a write lock on the object.  The caller can change
  // the lock state, make copies, etc., as desired.  The
  // GetNewSimulationState() routine scans through the list of states
  // held in the simulation_state array, looking for a pre-created state
  // that has no locks on it.  If one is found, a key is returned for
  // that state.  If all existing states in the simulation_state array
  // are locked, then a new state is created (subject to the max state
  // count obtained from ReserveSimulationStateRequest() calls).  In all
  // cases the returned key holds a write lock on the state.  The states
  // are actually deleted only inside the director Release() command,
  // after destruction of all Oxs_Ext objects.
  //   NOTE: The key passed in to GetNewSimulationState() is allowed to
  // hold a lock on an Oxs_SimState, *IF* that state was obtained from a
  // previous call to GetNewSimulationState().  On entry,
  // GetNewSimulationState() releases any lock currently held by the
  // import key, but then immediately tests if the corresponding
  // Oxs_SimState is then free.  If so, then GetNewSimulationState()
  // resets the key to the same object.  Therefore, if the chance is
  // good that the input key holds the sole lock on an object, then the
  // lock should *not* be released before calling
  // GetNewSimulationState().
  void ReserveSimulationStateRequest(UINT4m reserve_count);
  void GetNewSimulationState(Oxs_Key<Oxs_SimState>& write_key);

};

#endif // _OXS_DIRECTOR
