/* FILE: energy.cc                 -*-Mode: c++-*-
 *
 * Abstract energy class, derived from Oxs_Ext class.
 *
 */

#include "energy.h"

/* End includes */

void Oxs_Energy::FillCalcCountOutput(const Oxs_SimState& state)
{
  calc_count_output.cache.state_id = state.Id();
  calc_count_output.cache.value    = static_cast<REAL8m>(calc_count);
}

// Constructors
Oxs_Energy::Oxs_Energy
( const char* name,     // Child instance id
  Oxs_Director* newdtr  // App director
  ) : Oxs_Ext(name,newdtr),calc_count(0),
      calc_count_output(this,InstanceName(),"Calc count","",0,
			&Oxs_Energy::FillCalcCountOutput,newdtr),
      energy_sum_output(this,InstanceName(),"Energy","J",1,
                        &Oxs_Energy::UpdateStandardOutputs,newdtr),
      field_output(this,InstanceName(),"Field","A/m",1,
                        &Oxs_Energy::UpdateStandardOutputs,newdtr)
  // MS VC++ 6.0 requires fully qualified member names
  // The energy_sum_output and field_output initialization above
  // is potentially suspect because they require pointers to *this,
  // which is not yet fully constructed.  In actual fact, the *this
  // is not dereferenced in the output object constructors, so it
  // will work okay, but it is probably better if eventually
  // initialization of these objects gets moved to Init().
{}

Oxs_Energy::Oxs_Energy
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr    // MIF block argument string
  ) : Oxs_Ext(name,newdtr,safe_interp,argstr),calc_count(0),
      calc_count_output(this,InstanceName(),"Calc count","",0,
			&Oxs_Energy::FillCalcCountOutput,newdtr),
      energy_sum_output(this,InstanceName(),"Energy","J",1,
                        &Oxs_Energy::UpdateStandardOutputs,newdtr),
      field_output(this,InstanceName(),"Field","A/m",1,
                        &Oxs_Energy::UpdateStandardOutputs,newdtr)
  // MS VC++ 6.0 requires fully qualified member names
  // The energy_sum_output and field_output initialization above
  // is potentially suspect because they require pointers to *this,
  // which is not yet fully constructed.  In actual fact, the *this
  // is not dereferenced in the output object constructors, so it
  // will work okay, but it is probably better if eventually
  // initialization of these objects gets moved to Init().
{}

//Destructor
Oxs_Energy::~Oxs_Energy() {}

// Default problem initializer routine.  This sets up default
// output objects, so any child that redefines this function
// will probably want to embed a call to this Init() inside
// the child specific version.
BOOL Oxs_Energy::Init()
{
  calc_count=0;
  return 1;
}


// Standard output object update interface
void Oxs_Energy::UpdateStandardOutputs(const Oxs_SimState& state)
{
  if(state.Id()==0) { // Safety
    return;
  }

  Oxs_MeshValue<REAL8m> energy;
  Oxs_MeshValue<ThreeVector> field;

  const Oxs_MeshValue<REAL8m>* energy_used=NULL;
  const Oxs_MeshValue<ThreeVector>* field_used=NULL;

  GetEnergyAndField(state,energy,energy_used,field,field_used);
}

///////////////////////////////////////////////////////////////////
// Field and energy calculation function.  This should be
// "conceptually const."  The energy return is average energy
// density for the corresponding cell, in J/m^3.  The field
// is in A/m.
//   There are two GetEnergyAndField functions.  The first is
// a private, virtual member function of Oxs_Energy.  It takes
// as imports Oxs_MeshValue references into which to store
// the results.  The second is a non-virtual public member
// function that passes in Oxs_MeshValue references that *may*
// be used to store the results, and separate parameters that
// return references to where the results actually were stored.
// If output caching in enabled, the Oxs_Energy base class calls
// the virtual GetEnergyAndField function with references to the
// output cache Oxs_MeshValue objects.  Otherwise the import
// buffer space is used.
//    Note: The *_used are references to pointers, because C++ does
//          not support references to references.
void
Oxs_Energy::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy_provided,
 const Oxs_MeshValue<REAL8m>* &energy_used,
 Oxs_MeshValue<ThreeVector>& field_provided,
 const Oxs_MeshValue<ThreeVector>* &field_used
 )
{
  ++calc_count;

  Oxs_MeshValue<REAL8m>* energy_buffer = &energy_provided;
  Oxs_MeshValue<ThreeVector>* field_buffer = &field_provided;

  if(field_output.GetCacheRequestCount()>0) {
    field_output.cache.state_id=0;
    field_buffer = &field_output.cache.value;
  }
  energy_buffer->AdjustSize(state.mesh);
  field_buffer->AdjustSize(state.mesh);

  GetEnergyAndField(state,*energy_buffer,*field_buffer);

  if(field_output.GetCacheRequestCount()>0) {
    field_output.cache.state_id=state.Id();
  }
  if(energy_sum_output.GetCacheRequestCount()>0) {
    UINT4m size = state.mesh->Size();
    REAL8m energy_sum=0;
    const Oxs_Mesh* mesh = state.mesh;
    for(UINT4m i=0; i<size; ++i) {
      energy_sum += (*energy_buffer)[i] * mesh->Volume(i);
    }
    energy_sum_output.cache.state_id=state.Id();
    energy_sum_output.cache.value=energy_sum;
  }

  energy_used=energy_buffer;
  field_used=field_buffer;
}

