/* FILE: atlasscalarfieldinit.cc      -*-Mode: c++-*-
 *
 * Atlas scalar field initializer, derived from Oxs_ScalarFieldInit
 * class.
 *
 */

#include <string>

#include "oc.h"
#include "nb.h"

#include "util.h"
#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "atlas.h"
#include "atlasscalarfieldinit.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_AtlasScalarFieldInit);

/* End includes */

// Constructor
Oxs_AtlasScalarFieldInit::Oxs_AtlasScalarFieldInit(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp_, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_ScalarFieldInit(name,newdtr,safe_interp_,argstr)
{
  // Process arguments
  CheckInitValueParamCount("atlas",1);
  const string& atlasname = (*FindInitValue("atlas"))[0];
  Oxs_Ext* tempobj = director->FindExtObject(atlasname);
  if(tempobj==NULL) {
    string msg=string("Unable to identify Oxs_Atlas object with name ")
      + atlasname;
    throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_Atlas* atlas
    = dynamic_cast<Oxs_Atlas*>(tempobj);
  if(atlas==NULL) {
    string msg=string("Object ")
      + atlasname
      + string(" is not an Oxs_Atlas.");
    throw Oxs_Ext::Error(msg.c_str());
  }
  atlaskey.Set(atlas); // Dependency lock is held until *this is deleted.
  DeleteInitValue("atlas");

  // Optional "default_value"; default is 0.
  REAL8m default_value=0.0;
  const vector<string>* vsptr = FindInitValue("default_value");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("default_value",1);
    default_value
      = Nb_Atof((*FindInitValue("default_value"))[0].c_str());
    DeleteInitValue("default_value");
  }

  // Fill out value vector with default value
  unsigned int i;
  for(i=0;i<atlas->GetRegionCount();i++)
    values.push_back(default_value);

  // Get and fill value vector with non-default values
  vsptr = FindInitValue("values");
  if(vsptr==NULL || vsptr->size()<1) {
    string msg = "MIF input block for \""
      + string(InstanceName())
      + "\" missing specification for parameter \"values\"";
      throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_SplitList name_value;
  for(i=0;i<vsptr->size();i++) {
    name_value.Split((*vsptr)[i].c_str());
    if( name_value.Count() != 2) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Error in MIF input block for \"%s\";"
		  " values[%d] sub-list must be a 2 element list,"
		  " { region_name value }"
		  " (actual sub-list size: %d)",
		  InstanceName(),i,name_value.Count());
      throw Oxs_Ext::Error(buf);
    }
    INT4m region_id = atlas->GetRegionId(name_value[0]);
    if(region_id<0) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Error in MIF input block for \"%s\";"
		  " First entry in name_value[%d] sub-list, \"%s\","
		  " is not a known region in atlas \"%s\"",
		  InstanceName(),i,name_value[0],atlas->InstanceName());
      throw Oxs_Ext::Error(buf);
    }
    values[region_id] = Nb_Atof(name_value[1]);
  }
  DeleteInitValue("values");

  VerifyAllInitArgsUsed();
}

BOOL Oxs_AtlasScalarFieldInit::Init()
{ return 1; }

void 
Oxs_AtlasScalarFieldInit::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<REAL8m>& array) const
{
  // Fill array with values based on atlas regions
  UINT4m i;
  UINT4m size=mesh->Size();
  array.AdjustSize(mesh);
  const Oxs_Atlas* atlas = atlaskey.GetPtr();
  if(atlas->GetRegionCount() != values.size()) {
    char buf[1024];
    sprintf(buf,"Region count mismatch (%d != %d) in"
	    " Oxs_ScalarFieldInit::FillMeshValue of object %s;"
	    " Atlas %s region count changed after initialization.",
	    atlas->GetRegionCount(),int(values.size()),
	    InstanceName(),atlas->InstanceName());
  }
  ThreeVector loc;
  for(i=0;i<size;i++) {
    mesh->Center(i,loc);
    INT4m id = atlas->GetRegionId(loc);
    if(id<0) {
      char buf[1024];
      sprintf(buf,"Mesh point %d, (%g,%g,%g), outside atlas %s",
	      i,loc.x,loc.y,loc.z,atlas->InstanceName());
      throw Oxs_Ext::Error(buf);
    } else {
      array[i] = values[id];
    }
  }
}
