/* FILE: exchange6ngbr.cc            -*-Mode: c++-*-
 *
 * 6 neighbor exchange energy on rectangular mesh,
 * derived from Oxs_Energy class.
 *
 */

#include "atlas.h"
#include "nb.h"
#include "key.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "exchange6ngbr.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_Exchange6Ngbr);

/* End includes */


// Constructor
Oxs_Exchange6Ngbr::Oxs_Exchange6Ngbr(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr),
    A_size(0), A(NULL), mesh_id(0)
{
  // Process arguments
  CheckInitValueParamCount("atlas",1);
  const string& atlasname = (*FindInitValue("atlas"))[0];
  Oxs_Ext* tempobj = director->FindExtObject(atlasname);
  if(tempobj==NULL) {
    string msg=string("Unable to identify Oxs_Atlas object with name ")
      + atlasname;
    throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_Atlas* atlas
    = dynamic_cast<Oxs_Atlas*>(tempobj);
  if(atlas==NULL) {
    string msg=string("Object ")
      + atlasname
      + string(" is not an Oxs_Atlas.");
    throw Oxs_Ext::Error(msg.c_str());
  }
  atlaskey.Set(atlas); // Dependency lock is held until *this is deleted.
  DeleteInitValue("atlas");

  // Check for optional default_A parameter; default is 0.
  REAL8m default_A=0.0;
  const vector<string>* vsptr = FindInitValue("default_A");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("default_A",1);
    default_A
      = Nb_Atof((*FindInitValue("default_A"))[0].c_str());
    DeleteInitValue("default_A");
  }

  // Allocate A matrix.  Because raw pointers are used, a memory
  // leak will occur if an exception is thrown inside this constructor.
  A_size = atlas->GetRegionCount();
  if(A_size<1) {
    string msg=string("Oxs_Atlas object ")
      + atlasname
      + string(" must contain at least one region.");
    throw Oxs_Ext::Error(msg.c_str());
  }
  A = new REAL8m*[A_size];
  A[0] = new REAL8m[A_size*A_size];
  UINT4m i;
  for(i=1;i<A_size;i++) A[i] = A[i-1] + A_size;
  for(i=0;i<A_size*A_size;i++) A[0][i] = default_A;

  // Fill A matrix
  vsptr = FindInitValue("A");
  if(vsptr==NULL || vsptr->size()<1) {
    string msg = "MIF input block for \""
      + string(InstanceName())
      + "\" missing specification for parameter \"A\"";
      throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_SplitList pair_list;
  for(i=0;i<vsptr->size();i++) {
    pair_list.Split((*vsptr)[i].c_str());
    if( pair_list.Count() != 3) {
      char buf[512];
      Oc_Snprintf(buf,sizeof(buf),
		  "Error in MIF input block for \"%s\";"
		  " A[%d] sub-list must be a 3 element list,"
		  " { region-1 region-2 A-value }"
		  " (actual sub-list size: %d)",
		  InstanceName(),i,pair_list.Count());
      throw Oxs_Ext::Error(buf);
    }
    INT4m i1 = atlas->GetRegionId(pair_list[0]);
    INT4m i2 = atlas->GetRegionId(pair_list[1]);
    REAL8m Apair = Nb_Atof(pair_list[2]);
    if(i1<0) {
      char buf[512];
      Oc_Snprintf(buf,sizeof(buf),
		  "Error in MIF input block for \"%s\";"
		  " First entry in A[%d] sub-list, \"%s\","
		  " is not a known region in atlas \"%s\"",
		  InstanceName(),i,pair_list[0],atlas->InstanceName());
      throw Oxs_Ext::Error(buf);
    }
    if(i2<0) {
      char buf[512];
      Oc_Snprintf(buf,sizeof(buf),
		  "Error in MIF input block for \"%s\";"
		  " Second entry in A[%d] sub-list, \"%s\","
		  " is not a known region in atlas \"%s\"",
		  InstanceName(),i,pair_list[1],atlas->InstanceName());
      throw Oxs_Ext::Error(buf);
    }
    A[i1][i2]=Apair;
    A[i2][i1]=Apair; // A should be symmetric
  }
  DeleteInitValue("A");

  VerifyAllInitArgsUsed();
}

Oxs_Exchange6Ngbr::~Oxs_Exchange6Ngbr()
{
  if(A_size>0 && A!=NULL) {
    delete[] A[0];
    delete[] A;
  }
}

void Oxs_Exchange6Ngbr::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  // See if mesh and/or atlas has changed.
  if(mesh_id !=  state.mesh->Id() || !atlaskey.SameState()) {
    // Setup region mapping
    mesh_id = 0; // Safety
    const Oxs_Atlas* atlas = atlaskey.GetPtr();
    UINT4m size = state.mesh->Size();
    region_id.AdjustSize(state.mesh);
    ThreeVector location;
    for(UINT4m i=0;i<size;i++) {
      state.mesh->Center(i,location);
      if((region_id[i] = atlas->GetRegionId(location))<0) {
	string msg="Import mesh to Oxs_Exchange6Ngbr::GetEnergyAndField()"
	  " routine of object " +  string(InstanceName())
	  + string(" has points outside atlas ")
	  + string(atlas->InstanceName());
	throw Oxs_Ext::Error(msg.c_str());
      }
    }
    mesh_id = state.mesh->Id();
  }
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    string msg="Import mesh to Oxs_Exchange6Ngbr::GetEnergyAndField()"
      " routine of object " +  string(InstanceName())
      + " is not an Oxs_RectangularMesh object.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  REAL8m wgtx = -1.0/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = -1.0/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = -1.0/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  REAL8m hcoef = -2/MU0;

  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address
	ThreeVector base = spin[i];
	REAL8m Msii = Ms_inverse[i];
	if(Msii == 0.0) {
	  energy[i]=0.0;
	  field[i].Set(0.,0.,0.);
	  continue;
	}
	REAL8m* Arow = A[region_id[i]];
	ThreeVector sum(0.,0.,0.);
	if(z>0) {
	  UINT4m j = i-xydim;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgtz*(spin[j] - base);
	}
	if(y>0) {
	  UINT4m j = i-xdim;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgty*(spin[j] - base);
	}
	if(x>0) {
	  UINT4m j = i-1;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgtx*(spin[j] - base);
	}
	if(x<xdim-1) {
	  UINT4m j = i+1;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgtx*(spin[j] - base);
	}
	if(y<ydim-1) {
	  UINT4m j = i+xdim;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgty*(spin[j] - base);
	}
	if(z<zdim-1) {
	  UINT4m j = i+xydim;
	  REAL8m Apair = Arow[region_id[j]];
	  if(Ms_inverse[j]!=0.0) sum += Apair*wgtz*(spin[j]- base);
	}
	field[i] = (hcoef*Msii) * sum;
	energy[i] = (sum * base);
      }
    }
  }
}
