/* FILE: scriptatlas.cc                 -*-Mode: c++-*-
 *
 * Script atlas class, derived from OXS extension class.
 *
 */

#include <string>

#include "oc.h"
#include "nb.h"

#include "scriptatlas.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_ScriptAtlas);

/* End includes */


// Constructor
Oxs_ScriptAtlas::Oxs_ScriptAtlas
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp_, // Safe interpreter
  const char* argstr    // MIF block argument string
  ) : Oxs_Atlas(name,newdtr,safe_interp_,argstr),
      safe_interp(safe_interp_)
#if (TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION>0))
  ,objcmd_argc(0),objcmd(NULL)
#endif // Tcl version check
{
  // Process arguments.
  CheckInitValueParamCount("xrange",2);
  const vector<string>& xrange = *FindInitValue("xrange");
  CheckInitValueParamCount("yrange",2);
  const vector<string>& yrange = *FindInitValue("yrange");
  CheckInitValueParamCount("zrange",2);
  const vector<string>& zrange = *FindInitValue("zrange");
  bounding_box.Set(xrange[0],xrange[1],
		   yrange[0],yrange[1],
		   zrange[0],zrange[1]);
  DeleteInitValue("xrange");
  DeleteInitValue("yrange");
  DeleteInitValue("zrange");

  const vector<string>* regionlistptr = FindInitValue("regions");
  if(regionlistptr==NULL || regionlistptr->size()<1) {
    string msg="No regions specified in input block for object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  region_name_list = *regionlistptr;
  DeleteInitValue("regions");

  const vector<string>* scriptlistptr = FindInitValue("script");
  if(scriptlistptr==NULL) {
    string msg="No script specified in input block for object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  script_list = *scriptlistptr;
  DeleteInitValue("script");

  VerifyAllInitArgsUsed();

#if (TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION>0))
  // If eval obj is available, then initialize obj command structures
  objcmd_argc=static_cast<UINT4m>(script_list.size())+9;
  objcmd = new Tcl_Obj*[objcmd_argc];
  UINT4m iarg;
  for(iarg=0;iarg<script_list.size();iarg++) {
    objcmd[iarg] 
      = Tcl_NewStringObj(Oc_AutoBuf(script_list[iarg].c_str()),
			 static_cast<int>(script_list[iarg].size()));
  }
  objcmd[objcmd_argc-9] = Tcl_NewDoubleObj(0.); // Dummy value
  objcmd[objcmd_argc-8] = Tcl_NewDoubleObj(0.); // ""
  objcmd[objcmd_argc-7] = Tcl_NewDoubleObj(0.); // ""
  objcmd[objcmd_argc-6] = Tcl_NewDoubleObj(bounding_box.GetMinX());
  objcmd[objcmd_argc-5] = Tcl_NewDoubleObj(bounding_box.GetMinY());
  objcmd[objcmd_argc-4] = Tcl_NewDoubleObj(bounding_box.GetMinZ());
  objcmd[objcmd_argc-3] = Tcl_NewDoubleObj(bounding_box.GetMaxX());
  objcmd[objcmd_argc-2] = Tcl_NewDoubleObj(bounding_box.GetMaxY());
  objcmd[objcmd_argc-1] = Tcl_NewDoubleObj(bounding_box.GetMaxZ());
  for(iarg=0;iarg<objcmd_argc;iarg++) Tcl_IncrRefCount(objcmd[iarg]);
#endif // Tcl version check
}

Oxs_ScriptAtlas::~Oxs_ScriptAtlas()
{
#if (TCL_MAJOR_VERSION > 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION>0))
  if(objcmd!=NULL) {
    for(UINT4m i=0;i<objcmd_argc;i++) Tcl_DecrRefCount(objcmd[i]);
    delete[] objcmd;
  }
#endif // Tcl version check
}


INT4m Oxs_ScriptAtlas::GetRegionId(const ThreeVector& point) const
{ // Appends point & bounding box to "script" from the Specify
  // block, and evaluates the resulting command in the safe
  // interpreter.  The script should return the 0-based index
  // for the corresponding region, or -1 if none apply.
#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  // For Tcl eval string interface
  string script=script_list[0];
  for(UINT4m iarg=1;iarg<script_list.size();iarg++) {
    script += " ";
    script += script_list[iarg];
  }
  char cmdbuf[4096];
  Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
	      "%s %.17g %.17g %.17g"
	      " %.17g %.17g %.17g %.17g %.17g %.17g",
	      script.c_str(),point.x,point.y,point.z,
              bounding_box.GetMinX(),bounding_box.GetMinY(),
	      bounding_box.GetMinZ(),bounding_box.GetMaxX(),
	      bounding_box.GetMaxY(),bounding_box.GetMaxZ());
  if(Tcl_Eval(safe_interp,cmdbuf)!=TCL_OK) {
    string msg=string("Error detected in Oxs_ScriptAtlas object ")
      + string(InstanceName()) + string("; Tcl script error: ")
      + string(Tcl_GetStringResult(safe_interp));
    throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_SplitList veclist;
  veclist.Split(Tcl_GetStringResult(safe_interp));
  if(veclist.Count()!=1) {
    string msg
      = string("Error detected in Oxs_ScriptAtlas object ")
      + string(InstanceName())
      + string(" Return script value is not a single integer: ")
      + string(Tcl_GetStringResult(safe_interp));
    throw Oxs_Ext::Error(msg.c_str());
  }
  return atoi(veclist[0]);
#else // ! (Tcl version < 8.1)
  // Use Tcl eval obj interface
  Tcl_SetDoubleObj(objcmd[objcmd_argc-9],point.x);
  Tcl_SetDoubleObj(objcmd[objcmd_argc-8],point.y);
  Tcl_SetDoubleObj(objcmd[objcmd_argc-7],point.z);
  if(Tcl_EvalObjv(safe_interp,objcmd_argc,objcmd,TCL_EVAL_GLOBAL)
     !=TCL_OK) {
    string msg=string("Error detected in Oxs_ScriptAtlas object ")
      + string(InstanceName()) + string("; Tcl script error: ")
      + string(Tcl_GetStringResult(safe_interp));
    throw Oxs_Ext::Error(msg.c_str());
  }
  int objc;
  int result;
  Tcl_Obj** objv;
  if(Tcl_ListObjGetElements(NULL,Tcl_GetObjResult(safe_interp),
			    &objc,&objv) != TCL_OK
     || objc != 1
     || Tcl_GetIntFromObj(NULL,objv[0],&(result)) != TCL_OK
     ) {
    string msg
      = string("Error detected in Oxs_ScriptAtlas object ")
      + string(InstanceName())
      + string(" Return script value is not a single integer: ")
      + string(Tcl_GetStringResult(safe_interp));
    throw Oxs_Ext::Error(msg.c_str());
  }
  return result;
#endif // (Tcl version < 8.1)
}

INT4m Oxs_ScriptAtlas::GetRegionId(const string& name) const
{ // Given a region id string (name), returns
  // the corresponding region id index.  If
  // "name" is not included in the atlas, then
  // -1 is returned.
  UINT4m count = UINT4m(region_name_list.size());
  for(UINT4m i=0;i<count;i++) {
    if(region_name_list[i].compare(name)==0) return i;
  }
  return -1;
}

INT4m Oxs_ScriptAtlas::GetRegionName(UINT4m id,string& name) const
{ // Given an id number, fills in "name" with
  // the corresponding region id string.  Returns
  // 0 on success, 1 if id is invalid.
  if(id>=region_name_list.size()) return 1;
  name = region_name_list[id];
  return 0;
}
