/* FILE: uuanisotropy.cc            -*-Mode: c++-*-
 *
 * Uniform Uniaxial Anisotropy, derived from Oxs_Energy class.
 *
 */

#include "oc.h"
#include "nb.h"
#include "threevector.h"
#include "director.h"
#include "simstate.h"
#include "ext.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "uniaxialanisotropy.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_UniaxialAnisotropy);

/* End includes */


// Constructor
Oxs_UniaxialAnisotropy::Oxs_UniaxialAnisotropy(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr), mesh_id(0)
{
  // Process arguments
  // Note: Stroustrup specifies an auto_ptr .reset() member function,
  // but that is apparently not supported by some common STL's.  As
  // a workaround, create a temporary auto_ptr, and use operator=().
  auto_ptr<Oxs_ScalarFieldInit>
    foo(NewScalarFieldInit(safe_interp,"K1"));
  K1_init = foo;
  auto_ptr<Oxs_VectorFieldInit>
    bar(NewVectorFieldInit(safe_interp,"axis"));
  axis_init = bar;
  VerifyAllInitArgsUsed();
}

void Oxs_UniaxialAnisotropy::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  UINT4m size = state.mesh->Size();
  if(mesh_id != state.mesh->Id()) {
    // This is either the first pass through, or else mesh
    // has changed.
    mesh_id = 0;
    K1_init->FillMeshValue(state.mesh,K1);
    axis_init->FillMeshValue(state.mesh,axis);
    for(UINT4m i=0;i<size;i++) {
      // Check that axis is a unit vector:
      const REAL8m eps = 1e-14;
      if(fabs(axis[i].MagSq()-1)>eps) {
	string msg="Invalid initialization detected for object "
	  + string(InstanceName())
	  + ": Anisotropy axis isn't norm 1";
	throw Oxs_Ext::Error(msg.c_str());
      }
    }
    mesh_id = state.mesh->Id();
  }

  for(UINT4m i=0;i<size;++i) {
    REAL8m k = K1[i];
    REAL8m field_mult = (2.0/MU0)*k*Ms_inverse[i];
    if(field_mult==0.0) {
      energy[i]=0.0;
      field[i].Set(0.,0.,0.);
      continue;
    }
    if(k<=0) {
      // Easy plane (hard axis)
      REAL8m dot = axis[i]*spin[i];
      field[i] = (field_mult*dot) * axis[i];
      energy[i] = -k*dot*dot; // Easy plane is zero energy
    } else {
      // Easy axis case.  For improved accuracy, we want to report
      // energy as -k*(dot*dot-1), where dot = axis * spin.  But
      // dot*dot-1 suffers from bad loss of precision if spin is
      // nearly parallel to axis.  The are a couple of ways around
      // this.  Recall that both spin and axis are unit vectors.
      // Then from the cross product:
      //            (axis x spin)^2 = 1 - dot*dot
      // The cross product requires 6 mults and 3 adds, and
      // the norm squared takes 3 mult and 2 adds
      //            => 9 mults + 5 adds.
      // Another option is to use
      //            (axis - spin)^2 = 2*(1-dot) 
      //     so  1 - dot*dot = t*(2-t)
      //                where t = 0.5*(axis-spin)^2.
      // The op count here is 
      //            => 5 mults + 6 adds.
      // Another advantage to the second approach is you get 'dot',
      // as opposed to dot*dot, which saves a sqrt if dot is needed.
      // The downside is that if axis and spin are anti-parallel,
      // then you want to use (axis+spin)^2 rather than (axis-spin)^2.
      // Since we don't need a high accuracy 'dot', we just use
      // the cross product formulation.  (I did some single-spin test
      // runs and the performance of the two methods was about the
      // same.) -mjd, 28-Jan-2001
      ThreeVector temp = axis[i];
      REAL8m dot = temp*spin[i];
      field[i] = (field_mult*dot) * temp;
      temp ^= spin[i];
      energy[i] = k*temp.MagSq();
    }
  }
}



